import time


class CustomTimer(object):
    def __init__(self, transaction, name):
        if not hasattr(transaction, 'custom_timers'):
            transaction.custom_timers = {}
        self.transaction = transaction
        self.name = name
        self.start = None

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, type, value, traceback):
        self.transaction.custom_timers[self.name] = time.time() - self.start



from __future__ import unicode_literals

import six
import zmq
import uuid
from threading import Thread


class BaseTurret(object):
    """The base turret class

    :param hq_address str: the ip address of the HQ runing OCT
    :param hq_pub_port int: the port of the publish socket in the HQ
    :param hq_rc_port int: the port of the result collector in the HQ
    """

    READY = 'ready'
    RUNNING = 'Running'
    ABORTED = 'Aborted'
    INIT = 'Initialized'
    KILLED = 'Killed'

    def __init__(self, config, script_module, unique_id=None):

        self.config = config
        self.cannons = []
        self.script_module = script_module
        self.start_time = None
        self.run_loop = True
        self.start_loop = True
        self.already_responded = False
        self.uuid = unique_id or six.text_type(uuid.uuid4())
        self.commands = {}
        self.status = self.INIT

        self.setup_sockets()

        self.init_commands()

    def init_commands(self):
        """Initialize the commands dictionnary. This dict will be used when master send a command to interpret them
        """
        pass

    def setup_sockets(self):
        """Init and connect all the turrets
        """
        self.context = zmq.Context()

        self.poller = zmq.Poller()
        self.local_result = self.context.socket(zmq.PULL)
        self.local_result.bind("inproc://turret")

        self.master_publisher = self.context.socket(zmq.SUB)
        self.master_publisher.connect("tcp://{}:{}".format(self.config['hq_address'], self.config['hq_publisher']))
        self.master_publisher.setsockopt_string(zmq.SUBSCRIBE, '')
        self.master_publisher.setsockopt_string(zmq.SUBSCRIBE, self.uuid)

        self.result_collector = self.context.socket(zmq.PUSH)
        self.result_collector.connect("tcp://{}:{}".format(self.config['hq_address'], self.config['hq_rc']))

        self.poller.register(self.local_result, zmq.POLLIN)
        self.poller.register(self.master_publisher, zmq.POLLIN)

    def close_sockets(self):
        """Close all the sockets
        """
        self.local_result.close()
        self.master_publisher.close()
        self.result_collector.close()
        self.context.destroy()

    def build_status_message(self):
        data = {
            'turret': self.config['name'],
            'status': self.status,
            'uuid': self.uuid,
            'rampup': self.config['rampup'],
            'script': self.config['script'],
            'cannons': self.config['cannons']
        }
        return data

    def find_command(self, payload):
        """Execute the given command by searching it in the self.commands property.

        :param payload str: the dict containing the message from the master
        :return: True if the command exists, false if the payload does not contain a command or the command is not found
        :rtype: bool
        """
        if 'command' in payload:
            command = self.commands.get(payload['command'])
            return command
        print("The message does not contain a command")
        return None

    def start(self):
        """Start the turret and wait for the master to call the run method
        """
        raise NotImplementedError("Start method must be implemented")

    def send_result(self, result):
        """Send result to the result collector

        :param result dict: the result to send to the master
        :return: None
        """
        raise NotImplementedError("send_result error must be implemented")

    def run(self):
        """Main loop for the turret
        """
        raise NotImplementedError("run method must be implemented")


class BaseCannon(Thread):
    """The base cannon class, inherit from thread

    :param start_time int: the start_time of the test
    :param run_time int: the total run time for the script in second
    :param script_module: the module containing the test
    """

    def __init__(self, start_time, script_module, turret_uuid, context, config):
        super(BaseCannon, self).__init__()
        self.start_time = start_time
        self.script_module = script_module
        self.run_loop = True
        self.config = config

        self.result_socket = context.socket(zmq.PUSH)
        self.result_socket.connect("tcp://{}:{}".format(self.config['hq_address'], self.config['hq_rc']))

    def run(self):
        """The main run method for the cannon
        """
        raise NotImplementedError("run method must be implemented")


class BaseTransaction(object):
    """The base transaction class for writing tests
    """

    def __init__(self, config):
        self.config = config
        self.custom_timers = {}

    def setup(self):
        """This method will be call before the run method, use it to setup all needed datas.
        The setup time will not be included in the scriptrun_time
        """
        pass

    def run(self):
        """This is the main function that will be executed for the tests
        """
        pass

    def tear_down(self):
        """This method will be call once the run method has ended. Since the transaction instance will never be
        destroyed before the end of the test, use this method to clean and reset all variables, etc.
        The tear down time will not be included in the scriptrun_time
        """
        pass



import sys
import time
import json
import logging
import traceback

from oct_turrets.base import BaseTurret
from oct_turrets.cannon import Cannon

log = logging.getLogger(__name__)


class Turret(BaseTurret):
    """This class represent the classic turret for oct
    """

    def init_commands(self):
        """Initialize the basics commandes for the turret
        """
        self.commands['start'] = self.run
        self.commands['status_request'] = self.send_status
        self.commands['kill'] = self.kill

    def send_status(self, msg=None):
        """Reply to the master by sending the current status
        """
        if not self.already_responded or self.status != self.READY:
            reply = self.build_status_message()
            self.result_collector.send_json(reply)

    def send_result(self, result):
        """Update the results and send it to the hq
        """
        result['turret_name'] = self.config['name']
        self.result_collector.send_json(result)

    def start(self):
        """Start the turret and wait for the master to run the test
        """
        log.info("starting turret")
        self.status = self.READY
        while self.start_loop:
            try:
                socks = dict(self.poller.poll(2000))
                self.send_status()
                if self.master_publisher in socks:
                    payload = self.master_publisher.recv_string()
                    payload = json.loads(payload)
                    command = self.find_command(payload)
                    if command:
                        command(payload['msg'])
            except (Exception, KeyboardInterrupt):
                self.close_sockets()
                raise

    def run(self, msg=None):
        """The main run method
        """
        if self.status == self.RUNNING:
            return None

        log.info("Starting test for turret %s", self.uuid)

        self.already_responded = True
        self.start_time = time.time()
        self.start_loop = False
        self.status = self.RUNNING
        self.send_status()

        if 'rampup' in self.config:
            rampup = float(self.config.get('rampup', 0)) / float(self.config['cannons'])
        else:
            rampup = 0

        last_insert = 0

        if rampup > 0 and rampup < 1:
            timeout = rampup * 1000
        else:
            timeout = 1000

        try:
            while self.run_loop:
                if len(self.cannons) < self.config['cannons'] and time.time() - last_insert >= rampup:
                    cannon = Cannon(self.start_time, self.script_module, self.uuid, self.context, self.config)
                    cannon.daemon = True
                    self.cannons.append(cannon)
                    cannon.start()
                    last_insert = time.time()

                socks = dict(self.poller.poll(timeout))
                if self.master_publisher in socks:
                    data = self.master_publisher.recv_string()
                    data = json.loads(data)
                    if 'command' in data and data['command'] == 'stop':  # not managed, must break the loop
                        log.info("Exiting loop, premature stop")
                        self.run_loop = False
                        break
                    elif 'command' in data:
                        command = self.find_command(data)
                        if command:
                            command(data['msg'])
                if self.local_result in socks:
                    results = self.local_result.recv_json()
                    self.send_result(results)

            self.reset_turret()

        except (Exception, RuntimeError, KeyboardInterrupt) as e:
            self.status = self.ABORTED
            log.error(e)
            self.send_status()
            traceback.print_exc()
            self.close_sockets()

    def reset_turret(self):
        """Reset the turret and set it ready for the next test
        """
        log.info("Sending stop signal to cannons...")
        for i in self.cannons:
            i.run_loop = False
        log.info("Waiting for all cannons to finish")
        for i in self.cannons:
            i.join()

        self.cannons = []
        self.already_responded = False
        self.start_loop = True
        self.run_loop = True

        # clear sockets
        self.close_sockets()
        self.setup_sockets()

        self.start()

    def kill(self, msg=None):
        """Kill the turret
        """
        for i in self.cannons:
            i.run_loop = False
        self.status = self.KILLED
        self.send_status()
        self.close_sockets()
        sys.exit(1)



import os
import sys
import six
import uuid
import logging
import tarfile
import tempfile
import argparse
from oct_turrets.turret import Turret
from oct_turrets.exceptions import InvalidConfiguration, InvalidTarTurret
from oct_turrets.utils import validate_conf, clean_tar_tmp, load_file

log = logging.getLogger(__name__)


def unpack(turret_achive, dir_name):
    """Unpack the turret and add extract path to PYTHONPATH

    :param str unique_id: the unique turret identifier
    :return: False is there is a problem with the tarball, True otherwise
    :rtype: bool
    """
    if not tarfile.is_tarfile(turret_achive):
        return False
    with tarfile.open(turret_achive) as tar:
        tar.extractall(dir_name)
    sys.path.append(dir_name)
    return True


def from_tar(tar_path, unique_id, dir_name):
    """Load all require components from tar archive

    :param str tar_path: the path to json configuration file
    :param str unique_id: the unique identifier of the turret
    :param str dir_name: the full path for extraction
    :return: module file and loaded configuration
    """

    if not unpack(tar_path, dir_name):
        raise InvalidTarTurret("Invalide tar file provided")

    try:
        config = validate_conf(os.path.join(dir_name, 'config.json'))
    except InvalidConfiguration:
        clean_tar_tmp(dir_name, True)
        raise

    module_file = os.path.join(dir_name, config['script'])
    module = load_file(module_file)
    return module, config


def from_config(config_path):
    """Load all require components from configuration file

    :param str config_path: the path to json configuration file
    :return: module file and loaded configuration
    """
    config = validate_conf(config_path)
    cfg_path = os.path.dirname(config_path)
    module_file = os.path.join(cfg_path, config['script'])
    module = load_file(module_file)
    return module, config


def start(args):
    unique_id = six.text_type(uuid.uuid4())
    is_tar = None

    if args.tar is not None:
        dir_name = "{}-{}".format(os.path.basename(args.tar).split('.')[0], unique_id)
        dir_name = os.path.join(tempfile.gettempdir(), dir_name)
        module, config = from_tar(args.tar, unique_id, dir_name)
        is_tar = True
    else:
        module, config = from_config(args.config)

    turret = Turret(config, module, unique_id)

    try:
        turret.start()
    except (Exception, KeyboardInterrupt):
        clean_tar_tmp(dir_name, is_tar)
        raise

    clean_tar_tmp(dir_name, is_tar)


def main():

    parser = argparse.ArgumentParser(description='Give parameters for start a turret instance')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--config-file', type=str, default=None, help='path for json configuration file')
    group.add_argument('--tar', type=str, default=None, help='Path for the tarball to use')
    args = parser.parse_args()

    if args.config_file is None and args.tar is None:
        parser.error('You need a config_file.json to start a turret')

    start(args)



import os
import imp
import inspect
import json
import shutil
import tempfile

from oct_turrets.config import REQUIRED_CONFIG_KEYS
from oct_turrets.exceptions import InvalidConfiguration


def is_test_valid(test_module):
    """Test if the test_module file is valid

    We only check for the transaction class, since the generic transaction
    parent will raise a NotImplemented exception if the run method is not present
    :return: True if the test is valid, else False
    :rtype: bool
    :raises: InvalidTestError
    """
    if not hasattr(test_module, "Transaction"):
        raise Exception("No transaction class found in test script")
    getattr(test_module, "Transaction")
    return True


def load_module(path):
    """Load a single module based on a path on the system

    :param str path: the full path of the file to load as a module
    :return: the module imported
    :rtype: mixed
    :raises: ImportError, InvalidTestError
    """
    if not os.path.exists(path):
        raise ImportError("File does not exists: {}".format(path))
    try:
        module_name = inspect.getmodulename(os.path.basename(path))
        module = imp.load_source(module_name, path)
        if is_test_valid(module):
            return module
        raise Exception("Module not valid")
    except ImportError as e:
        raise Exception("Error importing the tests script {}\nError: {}".format(module_name, e))


def load_file(file_name):
    """Load a single file based on its full path as a module

    :param str filename: the full path of the file
    :return: the module loaded
    :rtype: mixed
    :raises: ImportError, InvalidTestError
    """
    if not os.path.exists(file_name):
        raise ImportError("File does not exists: {}".format(file_name))
    realpath = os.path.realpath(os.path.abspath(file_name))
    return load_module(realpath)


def validate_conf(config_file):
    """Check a configuration file in json format and return the loaded json.
    If a required key is not present in the file or if the file does not exists a InvalidConfiguration exception
    will be raised

    :param str config_file: the full config.json path
    :param str tar: the full tar file path
    :return: the loaded config object
    :rtype: dict
    :raise: InvalidConfig
    """

    # check if the file exists
    if not os.path.isfile(config_file):
        raise InvalidConfiguration("The given configuration file does not exist")

    with open(config_file) as f:
        data = json.load(f)

    for key in REQUIRED_CONFIG_KEYS:
        if key not in data:
            raise InvalidConfiguration("Missing required configuration key %s" % key)

    return data


def clean_tar_tmp(dir_name, is_tar):
    """This method will try to remove temp files from extracted tar
    """
    if not is_tar:
        return None
    try:
        shutil.rmtree(dir_name)
    except OSError:
        pass  # files already removed



import time

from oct_turrets.base import BaseCannon


class Cannon(BaseCannon):

    def run(self):
        """The main run method of the canon
        """
        elapsed = 0
        trans = self.script_module.Transaction(self.config)
        trans.custom_timers = {}

        while self.run_loop:
            error = ''

            trans.setup()
            start = time.time()
            try:
                trans.run()
            except Exception as e:
                error = str(e)
            scriptrun_time = time.time() - start
            trans.tear_down()

            elapsed = time.time() - self.start_time

            epoch = time.mktime(time.localtime())

            data = {
                'elapsed': elapsed,
                'epoch': epoch,
                'scriptrun_time': scriptrun_time,
                'error': error,
                'custom_timers': trans.custom_timers
            }
            self.result_socket.send_json(data)



from setuptools import setup

from oct_turrets import __version__

setup(
    name='oct-turrets',
    version=__version__,
    author='Emmanuel Valette',
    packages=['oct_turrets'],
    include_package_data=True,
    description='Client tester part for oct',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4'
    ],
    install_requires=[
        'pyzmq',
        'six'
    ],
    entry_points={'console_scripts': [
        'oct-turrets-start = oct_turrets.start_turret:main'
    ]},
)
